@@ -9,8 +9,9 @@ from random import Random
 from types import EllipsisType
 from typing import Annotated
 from typing import ClassVar
-from typing import NamedTuple
+from typing import Literal
 from typing import Protocol
+from typing import TypeGuard
 
 from docnote import ClcNote
 from typing_extensions import TypeIs
@@ -18,11 +19,7 @@ from typing_extensions import TypeIs
 if typing.TYPE_CHECKING:
     from _typeshed import DataclassInstance
 
-    from templatey.environments import AsyncTemplateLoader
-    from templatey.environments import SyncTemplateLoader
-    from templatey.templates import SegmentModifier
-    from templatey.templates import TemplateConfig
-    from templatey.templates import TemplateSignature
+    from templatey._signature import TemplateSignature
 else:
     DataclassInstance = object
 
@@ -37,9 +34,19 @@ class InterfaceAnnotationFlavor(Enum):
 
 @dataclass(frozen=True)
 class InterfaceAnnotation:
+    """
+    """
     flavor: InterfaceAnnotationFlavor
 
 
+class _TemplateClassSingleton(Enum):
+    DYNAMIC_TEMPLATE_CLASS = 'dynamic'
+type DynamicTemplateClass = Literal[
+    _TemplateClassSingleton.DYNAMIC_TEMPLATE_CLASS]
+DYNAMIC_TEMPLATE_CLASS: DynamicTemplateClass = \
+    _TemplateClassSingleton.DYNAMIC_TEMPLATE_CLASS
+
+
 # Technically this should be an intersection type with both the
 # _TemplateIntersectable from templates and the DataclassInstance returned by
 # the dataclass transform. Unfortunately, type intersections don't yet exist in
@@ -157,7 +164,7 @@ type DynamicClassSlot[T: TemplateParamsInstance] = Annotated[
         ''')]
 
 
-def is_template_class(cls: type) -> TypeIs[type[TemplateIntersectable]]:
+def is_template_class_xable(cls: type) -> TypeIs[type[TemplateIntersectable]]:
     """Rather than relying upon @runtime_checkable, which doesn't work
     with protocols with ClassVars, we implement our own custom checker
     here for narrowing the type against TemplateIntersectable. Note
@@ -165,14 +172,12 @@ def is_template_class(cls: type) -> TypeIs[type[TemplateIntersectable]]:
     re: the missing intersection type in python, though support might be
     unreliable depending on which type checker is in use.
     """
-    return (
-        hasattr(cls, '_templatey_config')
-        and hasattr(cls, '_templatey_resource_locator')
-        and hasattr(cls, '_templatey_signature')
-    )
+    return hasattr(cls, '_templatey_signature')
 
 
-def is_template_instance(instance: object) -> TypeIs[TemplateIntersectable]:
+def is_template_instance_xable(
+        obj: object
+        ) -> TypeIs[TemplateIntersectable]:
     """Rather than relying upon @runtime_checkable, which doesn't work
     with protocols with ClassVars, we implement our own custom checker
     here for narrowing the type against TemplateIntersectable. Note
@@ -180,29 +185,29 @@ def is_template_instance(instance: object) -> TypeIs[TemplateIntersectable]:
     re: the missing intersection type in python, though support might be
     unreliable depending on which type checker is in use.
     """
-    return is_template_class(type(instance))
+    return is_template_class_xable(type(obj))
+
+
+def is_template_class(cls: type) -> TypeGuard[TemplateClass]:
+    """Like is_template_class_xable, but for the raw template class.
+    """
+    return isinstance(cls, type) and hasattr(cls, '_templatey_signature')
+
+
+def is_template_instance(obj: object) -> TypeGuard[TemplateParamsInstance]:
+    """Like is_template_instance_xable, but for the raw template class.
+    """
+    return (not isinstance(obj, type)) and hasattr(obj, '_templatey_signature')
 
 
 class TemplateIntersectable(Protocol):
     """This is the actual template protocol, which we would
-    like to intersect with the TemplateParamsInstance, but cannot.
-    Primarily included for documentation.
+    like to intersect with the TemplateParamsInstance, but cannot
+    (because python currently lacks an intersection type).
+
+    Partly here for documentation, partly for use in ``cast`` calls.
     """
-    _templatey_config: ClassVar[TemplateConfig]
-    # Note: whatever kind of object this is, it needs to be understood by the
-    # template loader defined in the template environment. It would be nice for
-    # this to be a typvar, but python doesn't currently support typevars in
-    # classvars
-    _templatey_resource_locator: ClassVar[object]
     _templatey_signature: ClassVar[TemplateSignature]
-    # Oldschool here for performance reasons; otherwise this would be a dict.
-    # Field names match the field names from the params; the value is gathered
-    # from the metadata value on the field.
-    _templatey_prerenderers: ClassVar[NamedTuple]
-    _templatey_segment_modifiers: ClassVar[tuple[SegmentModifier]]
-    # Used primarily for libraries shipping redistributable templates
-    _templatey_explicit_loader: ClassVar[
-        AsyncTemplateLoader | SyncTemplateLoader | None]
 
 
 # Note: we don't need cryptographically secure IDs here, so let's preserve
@@ -218,7 +223,7 @@ def create_templatey_id() -> int:
     """Templatey IDs are unique identifiers (theoretically, absent
     birthday collisions) that we currently use in two places:
     ++  as a scope ID, which is used when defining templates in closures
-    ++  for giving slot tree nodes a unique reference target for
+    ++  for giving slot/prerender tree nodes a unique reference target for
         recursion loops while copying and merging, which is more robust
         than ``id(target)`` and can be transferred via dataclass field
         into cloned/copied/merged nodes.
